#!/bin/ksh
#
# write by	: oldmanpushcart@gmail.com
# date		: 2010-06-28
# version	: 2.02
#

#
# -------------------------------------------------------------------
#                               全局变量定义                           #
# -------------------------------------------------------------------
#

# 本地字符编码
# 主要是用来转换抓取到的html页面、输入的搜索关键词、下载的mp3文件名以及以后考虑的mp3文件信息等。
# 脚本运行时会在init函数中根据$LANG环境变量获取
# 默认值是UTF-8
ENCODE="UTF-8"


# 百度mp3关键词搜索页面的url地址
# 这里做成可配置的形式，以后有变更也方便修改
# word后边跟的是GBK下的urlencode编码
BAIDU_MP3_SEARCH_URL="http://mp3.baidu.com/m?f=ms&rf=idx&tn=baidump3&ct=134217728&word="


# 百度mp3风格关键词搜索页面的url地址
# 该地址是负责按歌曲的类型进行搜索，老歌、伤感、轻音乐等都可以从这里进行搜索
# word后边的关键词也是GBK下的urlencode编码
BAIDU_MP3_STYLE_SEARCH_URL="http://mp3.baidu.com/m?f=ms&tn=baidump3&ct=134217728&lf=&rn=&lm=-1&rf=idxdaohang&word="


# 日志存放目录, 存放脚本运行时产生的日志目录
BASE_LOG_DIR="logs"


# 下载存放目录，下载的mp3存放的地址
BASE_DOWNLOAD_DIR="downloads"


# 网页抓取超时时间, 控制了脚本在抓取百度网页的超时时间
# 单位 : 秒
CURL_TIMEOUT=60


# 下载超时时间，控制了脚本在下载时候的超时时间
# 单位 : 秒
WGET_TIMEOUT=60


# 最小文件大小
# 下载下来的mp3如果太小了一般是没有任何意义的，很可能只是一些铃声什么的。
# 默认值是3M 如果歌曲下下来小于2M，认为无意义
# 单位：byte
MIN_FILESIZE=1024*1024*2


# 最大文件大小
# 下载下来的mp3如果太大了，一般也是没任何意义的，很可能只是一些杂乱的音乐集合
# 默认值是10M 如果歌曲下下来大于10M，认为无意义
# 单位：byte
MAX_FILESIZE=1024*1024*10





#
# -------------------------------------------------------------------
#                               脚本函数定义                           #
# -------------------------------------------------------------------
#

#
# 下载文件, 这个函数负责下载文件
# 下载之前会先去检查一下本地上是否已经有同名的文件，如果有了将不会去下载
# 下载采用wget的方式实现，下载的进程号写入pid文件，这样脚本在捕捉到SIGINT信号时可以根据里面的pid去kill掉相应的进程
# 先下载到一个download文件(保存文件名.download)中，下载完成后再重命名回来。下载过程中如果出现错误，则删除download文件
# 下载之后会去检查下载的文件大小是否大于$MIN_FILESIZE, 如果小于则将下载文件删除
function download
# 参数说明
#	$1 : 总进程ID, 主要是用来配合控制当前wget进程的
#	$2 : 下载URL
#	$3 : 下载完成后保存的文件
#	$4 : 自定义下载超时时间, 如果没定义则取脚本默认值 $WGET_TIMEOUT
# 返回值
#	无
{
	# 检查参数是否齐全
	if [[ -z "$1" || -z "$2" || -z "$3" ]]; then
		log "some args is empty, please check it, stop download. parent_pid=$1, url=$2, save_file=$3, timeout=$4"
		return
	fi

	# 检查是否设置了超时参数，没设置就让它等于脚本默认值
	[[ $4 ]] && local timeout=$4 || local timeout=$WGET_TIMEOUT
	
	# 构造下载后保存的文件路径和文件名
	log "start downloading, url=$2"
	local download_file=$BASE_DOWNLOAD_DIR/$3
	
	# 检查文件是否已存在
	if [[ -f $download_file ]]; then
		log "$download_file allready exists. stop download $2"
		return
	fi
		
	# 检查脚本是否已经标记为终止了, 如果pid文件中有TERMINATED标记的，那么说明脚本已经停止，不再进行新的下载
	if [[ `grep TERMINATED=$1 pid$1.pid` ]]; then
		log "process mark TERMINATED, stop download $2"
		return
	fi
	
	# 这里调用wget来实现文件的下载
	mkdir -p $BASE_DOWNLOAD_DIR
	wget -T $timeout "$2" -O $download_file.download&
	local wget_pid=$!
	
	# 检查脚本终止的补偿机制
	# 如果在下载开始之前未检测到脚本已经被标记未终止，但是下载开始之后脚本却已经完成了终止的操作，
	# 那么此时这个wget进程将不会被kill掉，所以这里需要一个补偿机制，将不受脚本控制的wget子进程干掉
	touch pid$1.pid
	if [[ `grep TERMINATED=$1 pid$1.pid` ]]; then
		log "repay check process is mark TERMINATED, stop download $2"
		echo "TERMINATED=$1" >> "pid$1.pid"
		kill $wget_pid
		return
	fi
	
	# 这里将wget的子进程号写入pid文件，同时等待wget调用的完成
	# 为了防止阻塞掉，所以在批量下载的时候一般都是新开启一个脚本来下载的
	log "wget pid=$wget_pid"
	echo "wget_pid=$wget_pid" >> "pid$1.pid"
	wait $wget_pid
	if [[ $? != 0 ]]; then
		rm -f $download_file.download
		log "download failed, url=$2. delete tmp file $download_file.download"
	else
		local real_filesize=`ls -l $download_file.download | awk '{print $5}'`
		if [[ $real_filesize -lt $MIN_FILESIZE ]]; then
			rm $download_file.download
			log "file size $real_filezie less then $MIN_FILESIZE. delete tmp file $download_file.download"
		elif [[ $real_filezie -gt $MAX_FILESIZE ]]; then
			rm $download_file.download
			log "file size $real_filezie greatness then $MAX_FILESIZE. delete tmp file $download_file.download"
		else
			mv $download_file.download $download_file
			log "download successed, url=$2. saved file $download_file"
		fi
	fi
}

#
# 将字符串进行URL编码
# 作者：heiner.steven@odn.de
function urlencode
# 参数说明
#	$1 : 原字符串
# 返回值
#	转码后的字符串
{
	echo "$1" | awk '
		BEGIN {
			EOL = "%0A"		# "end of line" string (encoded)
			split ("1 2 3 4 5 6 7 8 9 A B C D E F", hextab, " ")
			hextab [0] = 0
			for ( i=1; i<=255; ++i ) 
				ord [ sprintf ("%c", i) "" ] = i + 0
			if ("'"$EncodeEOL"'" == "yes") 
				EncodeEOL = 1; 
			else 
				EncodeEOL = 0
		}
		{
			encoded = ""
			for ( i=1; i<=length ($0); ++i ) {
				c = substr ($0, i, 1)
				if ( c ~ /[a-zA-Z0-9.-]/ ) {
					encoded = encoded c		# safe character
				} else if ( c == " " ) {
					encoded = encoded "+"	# special handling
				} else {
					lo = ord [c] % 16
					hi = int (ord [c] / 16);
					encoded = encoded "%" hextab [hi] hextab [lo]
				}
			}
			if ( EncodeEOL ) {
				printf ("%s", encoded EOL)
			} else {
				print encoded
			}
		}
	'
}


#
# 记录日志信息
# 日志格式为[时间]<pid=进程号,fun=调用函数名>:日志内容
function log
# 参数说明
#	$1 : 日志内容
{
	local timestamp=`date "+%H:%M:%S"`
	local log_filename="baidu_mp3.log.`date +%Y-%m-%d`"
	mkdir -p $BASE_LOG_DIR
	echo "[$timestamp]<pid=$$,fun=${FUNCNAME[1]}>:""$1" >> $BASE_LOG_DIR/$log_filename
}



#
# 抓取HTML网页内容
function html
# 参数说明
#	$1 : 网页URL
#	$2 : 服务端字符编码格式，默认值为$ENCODE
#	$3 : 客户端字符编码格式，默认值为$ENCODE
# 返回值
#	抓取的HTML内容
{
	[[ $2 ]] && local from_encode=$2 || local from_encode=$ENCODE
	[[ $3 ]] && local to_encode=$3 || local to_encode=$ENCODE
	log "FORM_ENODE=$from_encode, TO_ENCODE=$to_encode, URL=$1"
	echo "$1" | sed s/[][]//g | xargs curl -m $CURL_TIMEOUT -s | iconv -f $from_encode -t $to_encode 2>/dev/null
}




#
# 百度MP3详情页面，下载地址解码
# 百度非常无聊，将下载地址用javascript经过一次简单的加密之后才能得到
# 这个函数的作用就是解析详情页面，然后解码出下载地址
function decode_url_from_baidu_mp3_detail
# 参数说明
#	$1 : 详情URL
# 返回值
#	解码出的下载地址
{

	# 检查参数是否完整
	if [[ -z $1 ]]; then
		log "url is empty, please check it. url=""$1"
		return
	fi

	local detail_url=`echo "$1" | awk -F "&tn=baidusg" '{print $1}'`"&tn=baidusg"
	local detail=`html "$detail_url" GBK $ENCODE | egrep -e "var (sertim|encurl|subulrs)"`
	local sertim=`echo $detail | awk '{print $4}' | sed s/,//g`
	local encurl=`echo $detail | awk '{print $11}' | sed s/\"//g`
	
	# 检查页面上捕获的两个关键变量,以这两个变量是否能正确获取来判断网页抓取是否正确
	if [[ -z $sertim || -z $encurl ]]; then
		log "sertim or encurl is empty, can't continue to download, stop it"
		return
	fi
	
	log "sertim=$sertim , encurl=$encurl"
	
	local len=`echo $encurl | wc -m`
	local key=`expr $sertim % 26`
	[[ $key -le 0 ]] && key=1
	
	local asc_arr1=(48 49 50 51 52 53 54 55 56 57 65 66 67 68 69 70 71 72 73 74 75 76 77 78 79 80 81 82 83 84 85 86 87 88 89 90 97 98 99 100 101 102 103 104 105 106 107 108 109 110 111 112 113 114 115 116 117 118 119 120 121 122)
	
	for (( i=0;i<=61;i++ ))
	do
		local asc_arr2[${asc_arr1[$i]}]=$i
	done
	
	for (( i=0;i<$len;i++ ))
	do
		local ch=`echo $encurl | cut -c $[$i+1]`
		if [[ `echo $ch | sed s/[A-Za-z0-9]//g` = "" ]]; then
			local ch_code=`echo $ch | tr -d "\n" | od -An -t dC | sed s/[[:space:]]//g | grep -v ^$`
			local pos=$[${asc_arr2[$ch_code]}-$key]
			[[ $pos -lt 0 ]] && pos=$[$pos+62]
			ch=`echo ${asc_arr1[$pos]}| awk '{printf "%c",$1}'`
		fi
		local decurl+=$ch
	done
	local real_url=`echo $decurl | cut -c1-$[$len-1]`
	log "real_url=$real_url"
	echo "$real_url"
}



#
# 百度MP3关键词搜索
# 这个函数只是负责将搜索关键词进行编码转化后去搜索用的
function search_result_for_baidu_mp3
# 参数说明
#	$1 : 搜索关键词
# 返回值
#	排序好的搜索结果列表, 搜索结果列表只是mp3的详情页面地址，并非最终下载地址
{
	log "KEYWORDS=$1"
	local word=`echo "$1" | iconv -f $ENCODE -t GBK`
	word=`urlencode "$word"`
	
	local sort_result=`parse_html_for_baidu_mp3_search "$BAIDU_MP3_SEARCH_URL$word"`
	log "sort_result=""$sort_result"
	
	echo "$sort_result"
}



#
# 百度MP3搜索处理函数
# 这个函数负责抓取搜索结果页面，提取出搜索出的相关信息，并且按照权值计算公式
#	大小(K) * 速度 / ( 排列顺序 * 0.8 )
# 计算出每个结果的权值，然后再进行排序
function parse_html_for_baidu_mp3_search
# 参数说明
#	$1 : 搜索URL
# 返回值
#	权值最高的搜索结果
{
	local count=-1
	local table_mark=0
	local song_list=""
	local space_mark="SPACE"
	local html_content=`html $1 GBK`
	for line in `echo "$html_content" | awk '{print $0}' | sed s/[[:space:]]//g`
	do
		if [[ -n `echo "$line" | grep class=tdn` ]]; then
			count=0
			table_mark=1
		fi
		
		# 这里提取关键信息 歌曲名、作者、专辑、文件大小、文件类型、下载速度
		if [[ $count -gt 0 ]]; then
			case $count in
				1)	local song_name=`echo "$line" | sed "s/<[^>]*>//g"`
					local detail_url=`echo "$line" | awk -F "href=\"|\"title" '{print $2}'`
					[[ -z $song_name ]] && song_name="unknow"
					[[ -z $detail_url ]] && detail_url="unknow"
				;;
				
				2)	local auth=`echo $line | sed "s/<[^>]*>//g" | sed "s/&nbsp;//g"`
					[[ -z $auth ]] && auth="unknow"
				;;
				
				3)	local special=`echo $line | sed "s/<[^>]*>//g" | sed "s/&nbsp;//g"`
					[[ -z $special ]] && special="unknow"
				;;
				

				7)	local file_size=`echo $line | sed "s/<[^>]*>//g"`
					[[ -z `echo $file_size | egrep [0-9]` ]] && file_size="0K"
				;;
				
				8)	local type=`echo $line | sed "s/<[^>]*>//g"`
					[[ -z $type ]] && type="unknow"
				;;
				
				9)	local speed=`echo $line | sed "s/<td.*img\/mp3\/d//g" | sed "s/\..*$//g"`
					[[ -z $speed ]] && speed="0"
				;;

			esac
			[[ $count -eq 10 ]] && local song_list="$song_list\n$song_name$space_mark$detail_url$space_mark$auth$space_mark$special$space_mark$file_size$space_mark$type$space_mark$speed"
		fi
		
		[[ $count -ge 0 ]] && count=$[$count+1]
		[[ table_mark -eq 1 && -n `echo $line | egrep -e "<\/table>"` ]] && break 1
	done
	
	
	local sort_song_list=""
	local index=1
	for line in `echo "$song_list" | awk -F "$space_mark" '{if($6=="mp3")print $0}' | sed s/[[:space:]]//g`
	do
		line=`echo $line | sed s/$space_mark/" "/g`
		local rate=`echo $line | awk '{print $5}' | sed s/[Mm]/\ 1024/g | sed s/[Kk]/\ 1/g`" "`echo $line | awk '{print $7}'`" "$index
		local file_size=`echo "$rate" | awk '{printf "%f\n",$1*$2*1024}' | awk -F "." '{print $1}'`
		if [[ $file_size -gt $MIN_FILESIZE && $file_size -lt $MAX_FILESIZE ]]; then
			# 这个0.8是哥肉体调试出来的, 内牛满面ing T_T
			rate=`echo "$rate" | awk '{printf "%f\n",$1*$2*$3/($4*0.8)}'`
			sort_song_list="$sort_song_list\n$line $rate"
			index=$[$index+1]
		else
			log "file_size=$file_size, less then $MIN_FILESIZE or greatness then $MAX_FILESIZE, give up."
		fi
	done
	
	# 根据权值排序，并且返回第一条
	echo "$sort_song_list" | sort -rnk8 | head -1
	
}


#
# 抓取百度百行榜的页面信息
# 百度比较恶心，排行榜中并不会直接给出音乐的下载地址，甚至连音乐的详情页面都没有，只有音乐的搜索地址。
# 所以只能得到搜索地址之后，再去访问搜索页面，最后再访问音乐详情页面，才能最终得到下载地址
# 这里只是取得排行中列出的搜索地址URL，还好这个没加密，要不我就晕掉了
function gettop
# 参数说明
#	$1 : 排行榜URL地址
# 返回值
#	百度mp3关键词搜索请求URL列表
{
	log "get top, url=""$1"
	local html_content=`html $1 GBK`
	local search_url_list=`echo "$html_content" | grep "class=\"search\"" | awk -F "href=\"" '{print $2}' | awk -F "\" class=" '{print $1}'`
	local total=`echo "$search_url_list" | wc -l`
	log "total need download: $total"
	for search_url in $search_url_list
	do
		local top=`parse_html_for_baidu_mp3_search "$search_url"`
		if [[ -z top ]]; then
			log "found nothing! search_url=$search_url"
			continue
		fi
		local top_url=`echo "$top" | awk '{print $2}'`
		local top_name=`echo "$top" | awk '{print $1}'`
		local top_type=`echo "$top" | awk '{print $6}'`
		local download_url=`decode_url_from_baidu_mp3_detail "$top_url"`
		
		# 这里直接调用脚本自身去下载，相当于开了子进程，这样就能同时下载了
		# 进程的管理是通过stop函数对pid文件进行操作实现的, 多次执行脚本会产生多个pid文件，文件之间不会相互干扰，确保了下载的正常运行
		sh $0 "download" $$ "$download_url" "$top_name.$top_type" 30 &
	done
	log "done"
}


#
# 根据歌曲分类来搜索歌曲
# 过程和百度排行是一样的，返回的也是搜索请求URL
function getstyle
# 参数说明
#	$1 : 分类搜索URL
# 返回值
#	百度mp3关键词搜索请求URL列表
{
	local html_content=`html $1 GBK`
	local data=`echo "$html_content" | grep "var data = \["`
	local style_search_url="http://mp3.baidu.com/m?rf=djtopic&tn=baidump3&ct=134217728&lm=-1&ie=utf-8&word="
	local total=`echo "$style_search_url" | wc -l`
	log "total need download: $total"
	
	echo "$data" | awk 'BEGIN{RS="},"}{print $0}' | sed s/^,//g | sed s/^.*[[]//g | awk -F "," '{print $1"+"$2}' | awk -F "\"" '{print $2"+"$4}' | egrep -ve "^\+$" | while read line;
	do
		line=`echo "$line" | iconv -f $ENCODE -t UTF-8 | sed s/[[:space:]]/%20/g`
		local search_url=$style_search_url"$line"
		local top=`parse_html_for_baidu_mp3_search "$search_url"`
		if [[ -z top ]]; then
			log "found nothing! search_url=$search_url"
			continue
		fi
		local top_url=`echo "$top" | awk '{print $2}'`
		local top_name=`echo "$top" | awk '{print $1}'`
		local top_type=`echo "$top" | awk '{print $6}'`
		local download_url=`decode_url_from_baidu_mp3_detail "$top_url"`
		
		# 这里直接调用脚本自身去下载，相当于开了子进程，这样就能同时下载了
		# 进程的管理是通过stop函数对pid文件进行操作实现的, 多次执行脚本会产生多个pid文件，文件之间不会相互干扰，确保了下载的正常运行
		sh $0 "download" $$ "$download_url" "$top_name.$top_type" 30 &
	done
	log "done"
}


#
# 脚本初始化
function init
#
{
	log "PID=$$"

	# 初始化本地字符编码
	local local_encode=`echo $LANG | awk -F "." '{print $2}'`
	[[ -n $local_encode ]] && ENCODE=$local_encode
	log "ENCODE=$ENCODE"
	
	# 检查系统中是否有curl
	curl --version >> /dev/null 2>&1
	if [[ ! $? -eq 0 ]]; then
		log "Can't found the curl"
	fi
	
	# 检查系统中是否有wget
	wget --version >> /dev/null 2>&1
	if [[ ! $? -eq 0 ]]; then
		log "Can't found the wget"
	fi
	
	# 检查系统中是否有iconv
	iconv --version >> /dev/null 2>&1
	if [[ ! $? -eq 0 ]]; then
		log "Can't found the iconv"
	fi
	
}


#
# 关闭函数,标记脚本已经关闭，由SIGINT信号激活
function stop
{
	log "start stop..."
	# make Terminated
	echo "TERMINATED=$$" >> "pid$$.pid"
	# I know do like that is stupid, but I can't control the job
	#ps -ef | grep -v grep | grep wget | awk '{print $2}' | xargs kill
	for wget_pid in `grep wget pid$$.pid | awk -F "=" '{print $2}'`
	do	
		ps -ef | grep -v grep | grep wget | awk '{if($2==PID)print $2}' PID=$wget_pid | xargs kill
	done
	rm -f "pid$$.pid"
	log "finished stop!"
	exit 0
}

# 捕捉CTRL+C信号
trap "stop" SIGINT









#
# -------------------------------------------------------------------
#                               脚本开始                              #
# -------------------------------------------------------------------
#


init

#
# 参数说明
#	top100
#	top500
#	rihan
#	download <musicname>
#	download <pid> <url> <filename> <timeout>
case $1 in
	
	"t"|"top")
		if [[ $# -ge 2 ]]; then
			case $2 in
				# TOP100
				"top100") gettop "http://list.mp3.baidu.com/top/top100.html" ;;
	
				# TOP500
				"top500") gettop "http://list.mp3.baidu.com/top/top500.html" ;;
	
				# 日韩
				"rihan"	) gettop "http://list.mp3.baidu.com/top/rihan.html" ;;
	
				# 欧美
				"oumei"	) gettop "http://list.mp3.baidu.com/top/oumei.html" ;;
				
				# 影视金曲
				"moive"	) gettop "http://list.mp3.baidu.com/top/movie.html" ;;
				
				# 摇滚
				"rock"	) gettop "http://list.mp3.baidu.com/top/yaogun.html";;
				
				# 无效的参数
				*)	msg="top option not found!"
					log "$msg"
					echo "$msg"
				;;
			esac
		else
			msg="miss top options"
			log "$msg"
			echo "$msg"
		fi
	;;
	
	
	# 按照风格搜索
	"s"|"style")
		if [[ $# -lt 2 ]]; then
			#缺少风格关键词
			msg="miss style keywords"
			log "$msg"
			echo "$msg"
		else
			style_word=`echo "$2" | iconv -f $ENCODE -t GBK`
			style_word=`urlencode "$style_word"`
			getstyle "$BAIDU_MP3_STYLE_SEARCH_URL""$style_word"
		fi
	;;
	
	
	# 下载
	"d"|"download")
		case $# in
			# 搜索并下载一首歌曲
			2)	search_result=`search_result_for_baidu_mp3 "$2"`
				if [[ -z $search_result ]]; then
					log "found nothing, keywords=""$2"
				else
					search_result_url=`echo "$search_result" | awk '{print $2}'`
					search_result_name=`echo "$search_result" | awk '{print $1}'`
					search_result_type=`echo "$search_result" | awk '{print $6}'`
					download_url=`decode_url_from_baidu_mp3_detail "$search_result_url"`
					download $$ "$download_url" "$search_result_name.$search_result_type" 30
				fi
			;;
			
			# 由shell负责调用，用于批量下载
			5)	download $2 "$3" "$4" $5 ;;
			
			# 无效的参数
			*)	msg="download args error!"
				log "$msg"
				echo "$msg"
				exit -1
			;;
		esac
	;;
	
	# 无效的参数
	*)	msg="option not found!"
		echo "$msg"
	;;
	
esac

# 正常结束后会清理掉pid文件
[[ -f "pid$$.pid" ]] && rm -f "pid$$.pid"
